/*
 * Copyright (c) 2016, Denis Butin and Stefan-Lukas Gazdag
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
*/

#include <inttypes.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>

#include "hfas_draft.h"

/*
 * We use adrs as if it is read from "left to right" as index 0 to 128, so the
 * layer address of 8 bits is to be found "left".
 *
 * Please note that setBits and getBits refer to the bits according to their
 * position within the address and therefore starts counting the bits of the
 * address from 1, not from 0, while TSTBIT, SETBIT and CLRBIT start
 * counting from 0.
 *
 * Please note that we are aware of the advantages of macros. We simply use
 * these methods to show the basic idea of address operations.
 *
 */

int
setBits(unsigned char *adrs, unsigned int position, const unsigned int n,
    unsigned char *bits) {
	/* returns 1 if something's wrong with the position or size */
	if (position < 1 || n < 1)
		return 1;
	position -= 1;
	for (unsigned int i = 0; i < n; i++) {
		if (TSTBIT(bits,i))
			SETBIT(adrs,(i+position));
		else
			CLRBIT(adrs,(i+position));
	}
	return 0;
}

void
getBits(unsigned char *adrs, unsigned int position, const unsigned int n,
    unsigned char *bits) {
	position -= 1;
	for (unsigned int i = 0; i < n; i++) {
		if (TSTBIT(adrs,(i+position)))
			SETBIT(bits,i);
		else
			CLRBIT(bits,i);
	}
}

int
setBlockBit(unsigned char *adrs, const unsigned int bit) {
	/*
	 * The block bit is used for l-trees and hash trees.
	 * returns 1 if it is an OTS hash address
	 *
	 * A single bit:
	 */

	/* If this is NOT an OTS address: */
	if (!TSTBIT(adrs,55)) { /* Was 78 -(02), is now 55 */
		/*
		 * We could check if the ltreeBit is set,
		 * but as we talk about the same bit
		 * to flip, let's just do it.
		 */
		if (bit == 1)
			SETBIT(adrs,126);
		else if (bit == 0)
			CLRBIT(adrs,126);
	} else
		return 1;
	return 0;
}

int
setChainAddress(unsigned char *adrs, const unsigned int bits) {
	/* returns 1 if this is not an OTS hash address */
	unsigned char address[2];
	address[0] = (unsigned char)((bits >> 8) & 0xFF);
	address[1] = (unsigned char)(bits & 0xFF);
	/*
	 * The chain address is used with OTS.
	 * 16 bits
	 */
	/* If this is an OTS address: */
	if (TSTBIT(adrs,55)) {
		/* Gotta set bits from position 97 to 112 */
		setBits(adrs, 97, 16, address); /* Was 104 (-02), is now 97 */
	}
	else
		return 1;
	return 0;
}

int
setHashAddress(unsigned char *adrs, const unsigned int bits) {
	/* returns 1 if this is not an OTS hash address */
	unsigned char address[1];
	address[0] = (unsigned char)(bits & 0xFF);
	/* If this is an OTS address: */
	if (TSTBIT(adrs,55)) {
		/* 8 bits, from position 113 */
		setBits(adrs, 113, 8, address); /* Was 120 (-02), is now 113 */
	} else
		return 1;
	return 0;
}

void
setKeyBit(unsigned char *adrs, const unsigned int bit) {
	/*
	 * The key bit is used with any structure, but in the non-common part.
	 * single bit:
	 */
	if (bit == 1)
		SETBIT(adrs,127);
	else if (bit == 0)
		CLRBIT(adrs,127);
}

int
setTreeHeight(unsigned char *adrs, const unsigned int bits) {
	/* returns 1 if this is an OTS hash address */
	unsigned char address[1];
	address[0] = (unsigned char)(bits & 0xFF);
	/* If this is NOT an OTS address: */
        if (!TSTBIT(adrs,55)) {
		/* 8 bits from position 89 */
		setBits(adrs, 89, 8, address); /* Was 95/105, is now 89 */
	} else
		return 1;
	return 0;
}

int
getTreeHeight(unsigned char *adrs, unsigned int *bits) {
	/* returns 1 if this is an OTS hash address */
	unsigned char address[1];
	unsigned int tmp_bits = 0;
	memset(address, 0, sizeof(unsigned char));
	/* If this is NOT an OTS address: */
	if (!TSTBIT(adrs,55)) {
		getBits(adrs, 89, 8, address);
		tmp_bits = (unsigned int)address[0];
		memcpy(bits, &tmp_bits, sizeof(unsigned int));
	} else
		return 1;
	return 0;
}

int
setTreeIndex(unsigned char *adrs, const uint32_t bits) {
	/* returns 1 if this is an OTS hash address */
	/* If this is NOT an OTS address: */
	if (!TSTBIT(adrs,55)) {
		/* 24 bits from position 97 */
		unsigned char address[3];
		memset(address, 0, 2 * sizeof(unsigned char));
		address[0] = (unsigned char)((bits >> 16) & 0xFF);
		address[1] = (unsigned char)((bits >> 8) & 0xFF);
		address[2] = (unsigned char)(bits & 0xFF);
		setBits(adrs, 97, 24, address); /* Was 103 (-02), is now 97 */
	} else
		return 1;
	return 0;
}

void
setOTSBit(unsigned char *adrs, const unsigned int bit) {
	/*
	 * The OTS bit is always used.
	 * A single bit:
	 */
	if(bit == 1)
		SETBIT(adrs,55);
	else if (bit == 0)
		CLRBIT(adrs,55);
	for (int i = 56; i < 128; i++)
		CLRBIT(adrs,i);
}

int
setOTSAddress(unsigned char *adrs, const uint32_t bits) {
	/* returns 1 if this is not an OTS hash address */
	unsigned char address[3];
	address[2] = (unsigned char)(bits & 0xFF);
	address[1] = (unsigned char)((bits >> 8) & 0xFF);
	address[0] = (unsigned char)((bits >> 16) & 0xFF);
	/* The OTS address is only needed for the actual OTS case. */
	if (TSTBIT(adrs,55)) {
		/* 24 bits from position 73 */
		setBits(adrs, 73, 24, address); /* Was 80 (-02), is now 73 */
	} else
		return 1;
	return 0;
}

int
setLTreeBit(unsigned char *adrs, const unsigned int bit) {
	/*
	 * returns 1 if this is an OTS hash address
	 * The l-tree bit differs between hash trees and l-trees.
	 */
	if (!TSTBIT(adrs,55)) {
		if(bit == 1)
			SETBIT(adrs,63);
		else if (bit == 0)
			CLRBIT(adrs,63);
	} else
		return 1;
	// Cleanup for the rest of the address
	for (int i = 64; i < 128; i++)
		CLRBIT(adrs,i);
	return 0;
}

int
setLTreeAddress(unsigned char *adrs, const uint32_t bits) {
	/*
	 * returns 1 if this is an OTS hash address
	 * returns 2 if this is a hash tree address
	 */
	unsigned char address[3];
	memset(address, 0, sizeof(address));
	address[2] = (unsigned char)(bits & 0xFF);
	address[1] = (unsigned char)((bits >> 8) & 0xFF);
	address[0] = (unsigned char)((bits >> 16) & 0xFF);
	/* If this is NOT an OTS address: */
	if (!TSTBIT(adrs,55)) {
		if (TSTBIT(adrs,63)) {
			/* 24 bits from position 65 */
			setBits(adrs, 65, 24, address); /* Was 81, is now 65 */
		} else
			return 2;
	} else
		return 1;
	return 0;
}

int
getTreeIndex(unsigned char *adrs, uint32_t *bits) {
	/* returns 1 if this is an OTS hash address */
	uint32_t tmp_bits = 0;
	/* The tree index is needed for hash trees and l-trees. */
	/* If this is NOT an OTS address: */
	if (!TSTBIT(adrs,55)) {
		/* 24 bits, from position 97 */
		unsigned char address[3];
		memset(address, 0, sizeof(address));
		getBits(adrs, 97, 24, address); /* Was 103 (-02), is now 97 */
		tmp_bits = (((uint32_t)address[0]) << 16)|
		    (((uint32_t)address[1]) << 8)|
		    (uint32_t)address[2];
		memcpy(bits, &tmp_bits, sizeof(uint32_t));
	} else
		return 1;
	return 0;
}

/* Only needed for XMSS^MT */
void
setLayerAddress(unsigned char *adrs, const unsigned int bits) {
	unsigned char address[1];
	memset(address, 0, sizeof(address));
	address[0] = (unsigned char)(bits & 0xFF);
	/* 8 bits, from position 1 */
	setBits(adrs, 1, 8, address); /* Was 47 (-02), is now 1 */
}

/* Only needed for XMSS^MT */
void
setTreeAddress(unsigned char *adrs, const unsigned long long int bits) {
	unsigned char address[5];
	memset(address, 0, sizeof(address));
	address[4] = (unsigned char)(bits & 0xFF);
	address[3] = (unsigned char)((bits >> 8) & 0xFF);
	address[2] = (unsigned char)((bits >> 16) & 0xFF);
	address[1] = (unsigned char)((bits >> 24) & 0xFF);
	address[0] = (unsigned char)((bits >> 32) & 0xFF);
	/* Setting tree address: 40 bits, from position 9 */
	setBits(adrs, 9, 40, address); /* Was 55 (-02), is now 9 */
}

