/*
 * Copyright (c) 2016, Denis Butin and Stefan-Lukas Gazdag
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
*/

#include <openssl/hmac.h>
#include <openssl/sha.h>

#include "chacha_draft.h"
#include "prg_draft.h"

/*
 * To stick to the "mathematical" notion of the draft we use the same
 * naming for the PRF / G.
 */

/* NOTE: The seed is to be used as key here! */
int
G(unsigned char *r, const unsigned int rlen, const unsigned char *seed,
    const unsigned char adrs[16], const unsigned int bytes)
{
	/*
	 * returns 1 if bytes doesn't match 32 or 64 bytes
	 * returns 2 if the output lenth of HMAC is not equal to 64 bytesa
	 */
	unsigned char nonce[12] = {0};

	/*
	 * Nonce is the first 12 bytes (most-significant according to draft)
	 * of the address.
	 */
	for (unsigned int i = 0; i < 12 ; i++)
		nonce[i] = adrs[i];

	if (bytes == 32) {
		/* The counter is the last four bytes of the address */
		uint32_t counter;
		counter = (((uint32_t)adrs[12]) << 24) |
		          (((uint32_t)adrs[13]) << 16) |
		          (((uint32_t)adrs[14]) <<  8) | adrs[15];
  		CRYPTO_chacha_20_keystream(r, rlen, seed, nonce, counter);
	} else if (bytes == 64) {
		/* Taken from Andreas Huelsing's implementation: */
		unsigned int length;
		HMAC(EVP_sha512(), seed, rlen, adrs, 16, r, &length);
		if (length != 64)
			return 2;
	} else {
		return 1;
	}

	return 0;
}

int
PRFm(unsigned char *r, unsigned int rlen, const unsigned char *msg,
    const unsigned int msglen, const unsigned char *key,
    const unsigned int bytes)
{
	/*
	 * returns 1 if bytes doesn't match 32 or 64 bytes
	 * returns 2 if the output lenth of HMAC is not equal to 64 bytes
	 * HMAC-SHA2-256 OR HMAC-SHA512 for the 32-byte or the 64-byte solution
	 */
	if (bytes == 32) {
		HMAC(EVP_sha256(), key, bytes, msg, msglen, r, &rlen);
	} else if (bytes == 64) {
		unsigned long long left = rlen;
		uint32_t counter = 0;
		unsigned char *c = (unsigned char*)&counter;
		unsigned int length;
		unsigned int i = 0;
		unsigned char tmp[64];

		while (left > 0) {
			HMAC(EVP_sha512(), key, bytes, c, 4, tmp, &length);
			if (length != 64)
				return 2;

			for (i = 0; (i < length) && (i < left); i++)
				r[rlen - left + i] = tmp[i];

			left -= length;
			counter++;
		}
	} else {
		return 1;
	}

	return 0;
}
