/*
 * Copyright (c) 2016, Denis Butin and Stefan-Lukas Gazdag
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
*/

#include "xmssmt_draft.h"

/*
 * Outputs the reduced secret key of the x^th XMSS tree on the y^th layer.
 * We use the existing XMSS SK structure to store the reduced SK, instead of
 * a new structure, which would in turn require a second version of treeHash.
 */
void
getXMSS_SK(xmss_sk *R, xmss_mt_sk *SK_MT, unsigned int x, unsigned int y)
{
	/* Find the array index of the relevant reduced key */
	int count = -1;
	for (unsigned int j = SK_MT->d ; j > y + 1; j--)
		count += (1 << (SK_MT->hmt - j * (SK_MT->hmt / SK_MT->d )));
	count += x + 1;

	/*
	 * We do not get the index from reduced_keys, since reduced keys don't
	 * have indexes. Make sure to obtain index from SK_MT at some stage!
	 * Same for SK_PRF.
	 */
	R->bytes = SK_MT->reduced_keys[count].bytes;
	R->para = SK_MT->reduced_keys[count].para;
	R->h = SK_MT->hmt / SK_MT->d;

	for (unsigned int i = 0 ; i < (unsigned int)((1 << (R->h+1)) - 1); i++){
		memcpy(R->nodes[i].node,
		    SK_MT->reduced_keys[count].nodes[i].node, R->para.m);
		R->nodes[i].height = SK_MT->reduced_keys[count].nodes[i].height;
	}
	memcpy(R->wots_keys, SK_MT->reduced_keys[count].wots_keys, R->para.len
			* R->para.m * (1 << R->h) * sizeof(*R->wots_keys));
}

/* Initialize the XMSS^MT secret key SK_MT */
int
xmss_mt_sk_init(xmss_mt_sk *SK_MT, wots_param *para, unsigned int hmt,
    unsigned int d)
{
	/*
	 * returns 1 if a problem with calloc occurred
	 * returns 2 if the height hmt is less than 2
	 * returns 3 if d does not divide height hmt withouth remainder
	 * returns 4 if the call to RAND_bytes for SK_PRF failed
	 * returns 5 if the call to RAND_bytes for seed failed
	 */
	if (hmt < 2)
		return 2;

	if (hmt % d != 0)
		return 3;

	SK_MT->d = d;		/* Initialise d */
	SK_MT->hmt = hmt;	/* Initialise hmt */
	SK_MT->idx_MT = 0;	/* Initialise global index */

	/* Initialise SK_PRF */
	if ((SK_MT->SK_PRF = calloc(para->m, sizeof(*SK_MT->SK_PRF))) == NULL)
		return 1;

	/* Assign random bytes and store return code */
	int rcprf = RAND_bytes(SK_MT->SK_PRF, para->m);
 	if (rcprf != 1)
		return 4;

	/* Initialise seed */
	if ((SK_MT->seed = calloc(para->n, sizeof(*SK_MT->seed))) == NULL)
		return 1;

	/* set SEED to a uniformly random n-byte string */
	unsigned int rcseed = RAND_bytes(SK_MT->seed, para->n);
	if (rcseed != 1)
		return 5;

	/*
	 * Initialise the XMSS reduced_keys in the XMSS^MT SK
	 * There are 2^(h - (i * h) / d) XMSS trees at layer i, i between 1 and
	 * d and there is one XMSS reduced key per XMSS tree
	 */
	unsigned int num = 0;
	for (unsigned int j = 1 ; j < d + 1; j++)
		num = num + (1 << (hmt - j * (hmt / d)));

	if ((SK_MT->reduced_keys = calloc(num, sizeof(*SK_MT->reduced_keys)))
	    == NULL)
		return 1;

	unsigned int count = 0;
	unsigned char adrs[16];
	memset(adrs, 0, 16 * sizeof(*adrs));
	for (int layer = d - 1 ; layer > -1 ; layer--) {
		setLayerAddress(adrs, (uint32_t) layer);
		for (unsigned long long int tree = 0 ; tree < 
		    (unsigned long long int) (1 << ((d - 1 - (unsigned int)
		    layer) * (hmt / d))); tree++) {
			setTreeAddress(adrs, tree);
			xmss_sk_init(&(SK_MT->reduced_keys[count]), para,
			    hmt / d);
			memcpy(SK_MT->reduced_keys[count].SK_PRF, SK_MT->SK_PRF,
			    para->m);
			/* Will now call treeHash to compute nodes */
			unsigned char root[para->m];
			memset(root, 0, para->m);
			treeHash(root, &(SK_MT->reduced_keys[count]), 0,
			    hmt / d, SK_MT->seed, adrs);
			count++;
		}
	}
	return 0;
}

void
xmss_mt_sk_free(xmss_mt_sk *SK_MT)
{
	unsigned int num = 0;

	free(SK_MT->SK_PRF);
	free(SK_MT->seed);

	for (unsigned int j = 1 ; j < SK_MT->d + 1; j++)
		num = num + (1 << (SK_MT->hmt - j * (SK_MT->hmt / SK_MT->d)));

	for (unsigned int i = 0 ; i < num ; i++)
		xmss_sk_free(&(SK_MT->reduced_keys[i]));

	free(SK_MT->reduced_keys);
}

int
xmss_mt_sig_init(xmss_mt_sig *Sig, wots_param *para, unsigned int h,
    unsigned int d)
{
	/*
	 * returns 1 if a problem with calloc occurred
	 * returns 2 if the initalization of xmss_sig failed
	 */

	Sig->idx_sig = 0;
	Sig->randomness = calloc(para->m, sizeof(*Sig->randomness));
	if (Sig->randomness == NULL)
		return 1;

	xmss_sig_init(&(Sig->bottom_reduced_sig), para, h / d);
	Sig->reduced_sigs = calloc((d - 1), sizeof(*Sig->reduced_sigs));
	if (Sig->reduced_sigs == NULL)
		return 1;

	for (unsigned int i = 0 ; i < (d - 1); i++) {
		if ((xmss_sig_init(&(Sig->reduced_sigs[i]), para, h / d)) == 1)
			return 2;
	}

	/*
	 * This works because we assume m = n. If m != n, then we need an
	 * alternative xmss_sig_init since then len_n != len (meaning sig_ots
	 * has a different length).
	 */
	return 0;
}

void
xmss_mt_sig_free(xmss_mt_sig *Sig, unsigned int d)
{
	free(Sig->randomness);
	xmss_sig_free(&(Sig->bottom_reduced_sig));
	for (unsigned int i = 0 ; i < (d - 1); i++)
		xmss_sig_free(&(Sig->reduced_sigs[i]));

	free(Sig->reduced_sigs);
}

/*
 * Algorithm 14: XMSSMT_genPK
 * Generate an XMSS^MT public key from an XMSS^MT private key
 */
int xmss_mt_genpk(unsigned char *pk, xmss_mt_sk *SK_MT)
{
	/*
	 * returns 1 if a problem with the initialization of xmss_pk occurred
	 * returns 2 if a problem with the initialization of xmss_sk occurred
	 */
	unsigned int n = SK_MT->reduced_keys[0].para.n;

	unsigned char adrs[16];
	memset(adrs, 0, 16 * sizeof(*adrs)); /* toByte(adrs, 0, 16); */

	unsigned int d = SK_MT->d;
	setLayerAddress(adrs, d - 1);	/* ADRS.setLayerAddress(d-1) */

	/*
	 * We will return a string for the PK, but work with structures
	 * internally.
	 */
	xmss_pk PK;	/* Formats of XMSS_MT PK and XMSS PK are identical */

	/* Initialise the XMSS_MT PK */
	if ((xmss_pk_init(&PK, n)) == 1)
		return 1;

	xmss_sk R;	/* To store a reduced XMSS SK extracted from SK_MT */
	if (xmss_sk_init(&R, &(SK_MT->reduced_keys[0].para), SK_MT->hmt / d)
	    == 1)
		return 2;

	getXMSS_SK(&R, SK_MT, 0, d - 1); /* Store reduced secret key in R */

	/* PK_MT = root || SEED */
	memcpy(PK.seed, SK_MT->seed, n * sizeof(*PK.seed));

	unsigned char root[n];	/* This will store the result of treeHash */
	memset(root, 0, n * sizeof(*root));
	treeHash(root, &R, 0, SK_MT->hmt / d, PK.seed, adrs); /* Compute root */

	/* PK_MT = root || SEED */
	memcpy(PK.root, root, n * sizeof(*PK.root));

	/* Can use legacy function since same format as XMSS PK */
	xmssPKtoString(pk, &PK, n);

	/* Free the struct, since we end up working with strings */
	xmss_pk_free(&PK);
	xmss_sk_free(&R);

	return 0;
}

/*
 * XMSS_sign_wo_hash: like algorithm 11 (xmss_signing), but without
 * the initial message hash. Therefore, the values r (PRFm output) and sk_prf
 * are no longer needed.
 */
int
xmss_signing_wo_hash(xmss_sk *SK, unsigned char *msg, unsigned char *sig,
    unsigned char *seed, unsigned char adrs[16])
{
	/*
	 * returns 1 if the initialization of xmss_sig failed
	 * returns 2 if the index idx_sig is too large for the tree
	 * returns 3 if the access to SK->wots_keys would be out of bounds
	 */
	wots_param para = SK->para;
	xmss_sig Sig;
	unsigned int h = SK->h;
	if ((xmss_sig_init(&Sig, &para, h)) == 1)
		return 1;

	xmssStringToSig(&Sig, sig, para.m, SK->bytes, h, para.w);
	uint32_t idx_sig = SK->idx;
	if (idx_sig >= (uint32_t)(1 << h))
		return 2;

	unsigned int len = SK->para.len;
	unsigned int m = para.m;
	unsigned char auth[h * m];
	memset(auth, 0, sizeof(auth));
	buildAuth(SK, auth, idx_sig);
	setOTSBit(adrs, 1);
	setOTSAddress(adrs, idx_sig);
	unsigned char sig_ots[len * m];
	memset(sig_ots, 0, sizeof(sig_ots));
	if (idx_sig > (uint32_t)((1 << h) - 1))
		return 3;

	/* In the initial algorithm, next line uses tmp_msg instead of msg */
	genWotsSig(&SK->wots_keys[idx_sig * m * len], &SK->para, msg, sig_ots,
			seed, adrs);
	/* Pseudo: Sig = (idx_sig || sig_ots || auth); (no randomness r) */
	memcpy(Sig.sig_ots, sig_ots, len * m * sizeof(*Sig.sig_ots));
	memcpy(Sig.auth, auth, h * m * sizeof(*Sig.auth));
	Sig.idx_sig = SK->idx;
	SK->idx = idx_sig + 1;
	xmssSigToString(sig, &Sig, para.m, m, h, para.w);
	xmss_sig_free(&Sig);

	return 0;
}

/*
 * Extract the n most significant bits from source, store them in extracted.
 * Source is assumed to have length h bits -- this is used for idx_sig
 */
int
getMSB(uint64_t source, uint64_t *extracted, unsigned int n, unsigned int h)
{
	/*
	 * returns 1 if one tries to extract 0 bits
	 * returns 2 if one tries do extract more bits than possible
	 * returns 3 if one tries to extract more than 64 bits
	 */

	if (n == 0)
		return 1;

	if (n > h)
		return 2;

	if (h > 64)
		return 3;

	*extracted = source << (64 - h);
	*extracted = *extracted >> (64 - n);

	return 0;
}

/* Like getMSB, but for the n least significant bits. */
int
getLSB(uint64_t source, uint64_t *extracted, unsigned int n, unsigned int h)
{
	/*
	 * returns 1 if one tries to extract 0 bits
	 * returns 2 if one tries do extract more bits than possible
	 * returns 3 if one tries to extract more than 64 bits
	 */

	if (n == 0)
		return 1;

	if (n > h)
		return 2;

	if (h > 64)
		return 3;

	*extracted = source << (64 - n);
	*extracted = *extracted >> (64 - n);

	return 0;
}

/*
 * Algorithm 15: XMSSMT_sign
 * Generate an XMSS^MT signature and update the XMSS^MT secret key
 */
int
xmss_mt_sign(xmss_mt_sk *SK_MT, unsigned char *msg, unsigned int msglen,
    unsigned char *sig)
{
	/*
	 * returns 1 if the index is too large for the tree
	 * returns 2 if the initialization of xmss_mt_sig failed
	 * returns 3 if a problem with calloc occurred
	 * returns 4 if the initialization of xmss_sig failed
	 */
	unsigned int m = SK_MT->reduced_keys[0].bytes;
	unsigned char adrs[16], SK_PRF[m];
	memset(adrs, 0, sizeof(adrs)); /* ADRS = toByte(0,16) */
	/* SK_PRF = getSK_PRF(SK_MT) */
	memcpy(SK_PRF, SK_MT->SK_PRF, sizeof(SK_PRF));

	uint64_t idx_sig = SK_MT->idx_MT;	/* idx_sig = getIdx(SK_MT) */
	unsigned int h = SK_MT->hmt, d = SK_MT->d;

	if (idx_sig >= (uint64_t)(1 << h))
		return 1;

	/*
	 * if ((idx_sig + 1) == (uint64_t)(1 << h)) {
	 * // One might want to be informed that this is the last message
	 * // to be signed
	 * }
	 */

	SK_MT->idx_MT = idx_sig + 1;	/* setIdx(SK_MT, idx_sig + 1) */

	wots_param para = SK_MT->reduced_keys[0].para;
	xmss_mt_sig Sig_MT;
	if ((xmss_mt_sig_init(&Sig_MT, &para, h, d)) == 1)
		return 2;

	Sig_MT.idx_sig = idx_sig;	/* Sig_MT = idx_sig; */

	/* idx_tree = (h - h / d) most significant bits of idx_sig */
	uint64_t idx_tree = 0, idx_leaf = 0;
	getMSB(idx_sig, &idx_tree, h - h / d, h);

	/* idx_leaf = (h / d) least significant bits of idx_sig; */
	getLSB(idx_sig, &idx_leaf, h / d, h);

	/* SK = idx_leaf || SK_PRF || getXMSS_SK(SK_MT, idx_tree, 0) */
	xmss_sk SK;
	xmss_sk_init(&SK, &para, h / d);
	getXMSS_SK(&SK, SK_MT, idx_tree, 0);
	SK.idx = idx_leaf;
	memcpy(SK.SK_PRF, SK_PRF, m * sizeof(*SK.SK_PRF));

	setLayerAddress(adrs, 0); /* ADRS.setLayerAddress(0) */
	setTreeAddress(adrs, idx_tree); /* ADRS.setTreeAddress(idx_tree) */
	
	/* Sig_tmp = XMSS_signwo_hash(M', SK, SEED, ADRS) */
	unsigned char *Sig_tmp_string;
	if ((Sig_tmp_string = calloc(4 + para.m + para.len * para.m + (h / d) *
	    para.m, sizeof(unsigned char))) == NULL)
		return 3;

	unsigned char r[m];
	memset(r, 0, sizeof(r));
	unsigned char msg_tmp[m];
	memset(msg_tmp, 0, sizeof(msg_tmp));
	unsigned char sk_prf[m];
	memset(sk_prf, 0, m * sizeof(unsigned char));
	getSK_PRF(&SK, sk_prf);
	PRFm(r, m, msg, msglen, sk_prf, m); /* byte[m] */
	unsigned char input_tmp[2 * m];
	memset(input_tmp, 0, 2 * m * sizeof(unsigned char));
	unsigned char tobyte_tmp[m];
	memset(tobyte_tmp, 0, m * sizeof(unsigned char));
	toByte(tobyte_tmp, idx_sig, m);
	concat_bytes(input_tmp, tobyte_tmp, m, r, m);
	Hm(msg_tmp, msg, msglen, input_tmp, m); /* byte[m] */
	xmss_signing_wo_hash(&SK, msg_tmp, Sig_tmp_string, SK_MT->seed, adrs);
	xmss_sig Sig_tmp;
	if (xmss_sig_init(&Sig_tmp, &para, h / d) == 1)
		return 4;

	xmssStringToSig(&Sig_tmp, Sig_tmp_string, m, para.n, h / d, para.w);

	/* Sig_MT = Sig_MT | r | Sig_tmp */
	/*
	 * Sig_MT.bottom_reduced_sig = Sig_tmp;
	 * Reduced signatures contain no index and randomness,
	 * so no need to copy them
	 */
	memcpy(Sig_MT.bottom_reduced_sig.sig_ots, Sig_tmp.sig_ots,
	    para.len * m * sizeof(unsigned char));
	memcpy(Sig_MT.bottom_reduced_sig.auth, Sig_tmp.auth,
	    m * (h / d) * sizeof(unsigned char));
	memcpy(Sig_MT.randomness, r, m * sizeof(unsigned char));

	unsigned char root[m];

	for (unsigned int j = 1; j < d; j++) {
		treeHash(root, &SK, 0, h / d, SK_MT->seed, adrs);

		/* idx_leaf = (h / d) least significant bits of idx_tree; */
		getLSB(idx_tree, &idx_leaf, h / d, h);

		/*
		 * idx_tree = (h - j * (h / d))
		 * most significant bits of idx_tree
		 */
		getMSB(idx_tree, &idx_tree, h - j * (h / d), h);

		/* SK = idx_leaf || SK_PRF || getXMSS_SK(SK_MT, idx_tree, j); */
		getXMSS_SK(&SK, SK_MT, idx_tree, j);
		SK.idx = idx_leaf;
		memcpy(SK.SK_PRF, SK_PRF, m * sizeof(*SK.SK_PRF));

		setLayerAddress(adrs, j);	/* ADRS.setLayerAddress(j) */
		setTreeAddress(adrs, idx_tree);	/* ADRS.setTreeAddress(idx_tree) */

		/* Sig_tmp = XMSS_sign_wo_hash(root, SK, SEED, ADRS) */
		xmss_signing_wo_hash(&SK, root, Sig_tmp_string, SK_MT->seed,
		    adrs);

		/* Sig_MT = Sig_MT || Sig_tmp; */
		xmssStringToSig(&Sig_tmp, Sig_tmp_string, m, para.n, h / d,
		    para.w); /* convert back to struct */
		memcpy(Sig_MT.reduced_sigs[j-1].sig_ots, Sig_tmp.sig_ots,
		    para.len * m * sizeof(unsigned char));
		memcpy(Sig_MT.reduced_sigs[j-1].auth, Sig_tmp.auth,
		    (h / d) * m * sizeof(unsigned char));
	}
	xmss_mt_SigToString(sig, &Sig_MT, m, m, h, para.w, d);

	xmss_sk_free(&SK);
	xmss_mt_sig_free(&Sig_MT, d);
	xmss_sig_free(&Sig_tmp);
	free(Sig_tmp_string);

	return 0;
}

/*
 * XMSS_rootFromSig_wo_hash: like algorithm 12 (xmss_rootFromSig),
 * but without the initial message hash. r is no longer needed.
 */
int
xmss_rootFromSig_wo_hash(xmss_sig *Sig, unsigned char *msg, unsigned char *node,
    unsigned char *seed, unsigned int m, unsigned int h, unsigned char adrs[16])
{
	/*
	 * returns 1 if the index is too large for the tree
	 * Note: what we originally expected as msglen is now simply m
	 * If you actually get it from Sig: the last h-1 * n bytes
	 */
	unsigned char auth[h * m];
	memset(auth, 0, sizeof(auth));
	memcpy(auth, Sig->auth, sizeof(auth));

	/* If you actually get it from Sig: bytes 0 - 3 */
	uint32_t idx_sig = Sig->idx_sig;
	if (idx_sig >= (uint32_t)(1 << h))
		return 1;

	wots_param para;
	para.m = m;
	para.w = 16;
	para.len_1 = calcLen1(para.w, para.m);
	para.len_2 = calcLen2(para.w, para.len_1);
	para.len = calcLen(para.len_1, para.len_2);
	unsigned char sig_ots[para.len * m];
	memset(sig_ots, 0, sizeof(sig_ots));
	memcpy(sig_ots, Sig->sig_ots, sizeof(sig_ots));
	setOTSBit(adrs, 1);
	setOTSAddress(adrs, idx_sig);
	unsigned char pk_ots[para.len * m];
	memset(pk_ots, 0, sizeof(pk_ots));

	/* msg_tmp instead of msg in initial algo */
	wotsPubKeyFromSig(&para, msg, sig_ots, pk_ots, seed, adrs);
	setOTSBit(adrs, 0);
	setLTreeBit(adrs, 1);
	setLTreeAddress(adrs, idx_sig);
	unsigned char node0[para.m], node1[para.m];
	memset(node0, 0, sizeof(node0));
	memset(node1, 0, sizeof(node1));
	ltree(pk_ots, node0, &para, seed, adrs);
	setLTreeBit(adrs, 0);
	setTreeIndex(adrs, idx_sig);
	unsigned int index = 0;
	for (unsigned int k = 0; k < h; k++) {
		setTreeHeight(adrs, k);

		if (((int)floor((double)((idx_sig / (1 << k)) % 2))) == 0) {
			getTreeIndex(adrs, &index);
			setTreeIndex(adrs, (index >> 1));
			rand_hash(node1, node0, &auth[k * m], seed, adrs, m);
		} else {
			getTreeIndex(adrs, &index);
			setTreeIndex(adrs, ((index - 1) >> 1));
			rand_hash(node1, &auth[k * m], node0, seed, adrs, m);
		}
		memcpy(node0, node1, sizeof(node0));
	}
	memcpy(node, node0, sizeof(node0));

	return 0;
}

/*
 * Algorithm 16: XMSSMT_verify
 * Verify an XMSS^MT signature Sig_MT on a message M using an XMSS^MT
 * public key PK_MT
 */
int
xmss_mt_verify(unsigned char *pk_all, unsigned char *sig, unsigned char *msg,
    unsigned int msglen, wots_param *para, unsigned int h, unsigned int d)
{
	/*
	 * returns 0 if verification was successful
	 * returns 1 if verification was NOT successful
	 * returns 2 if the initialization of xmss_mt_sig failed
	 * returns 3 if the initialization of xmss_pk failed
	 */
	xmss_mt_sig Sig_MT;
	if ((xmss_mt_sig_init(&Sig_MT, para, h, d)) == 1)
		return 2;

	unsigned int m = para->m, w = para->w, n = para->n;
	xmss_mt_StringToSig(&Sig_MT, sig, m, n, h, w, d);

	uint64_t idx = Sig_MT.idx_sig;	/* idx = getIdx(Sig_MT) */

	xmss_pk PK_MT;	/* Use existing XMSS PK structure */
	if ((xmss_pk_init(&PK_MT, m)) == 1)
		return 3;

	xmssStringToPK(&PK_MT, pk_all, m);

	unsigned char seed[n], adrs[16];
	/* SEED = getSEED(PK_MT) */
	memcpy(seed, PK_MT.seed, sizeof(seed));
	memset(adrs, 0, sizeof(adrs));	/* ADRS = toByte(0,16) */
	
	uint64_t idx_leaf = 0, idx_tree = 0;
	/* idx_leaf = (h / d) least significant bits of idx; */
	getLSB(idx, &idx_leaf, h / d, h);
	/* idx_tree = (h - h / d) most significant bits of idx; */
	getMSB(idx, &idx_tree, h - h / d, h);
	
	xmss_sig Sigp; /* Sig' */
	xmss_sig_init(&Sigp, para, h / d);

	/* Sig' = idx_leaf || getR(Sig_MT) || getXMSSSignature(Sig_MT, 0) */
	Sigp.idx_sig = idx_leaf;
	memcpy(Sigp.randomness, Sig_MT.randomness,
	    m * sizeof(*Sigp.randomness));
	memcpy(Sigp.sig_ots, Sig_MT.bottom_reduced_sig.sig_ots,
	    para->len * m * sizeof(*Sigp.sig_ots));
	memcpy(Sigp.auth, Sig_MT.bottom_reduced_sig.auth,
	    (h / d) * m * sizeof(*Sigp.auth));

	setLayerAddress(adrs, 0);	/* ADRS.setLayerAddress(0) */
	setTreeAddress(adrs, idx_tree);	/* ADRS.setTreeAddress(idx_tree) */

	unsigned char r[m];
	memset(r, 0, sizeof(r));
	memcpy(r, Sig_MT.randomness,
	    m * sizeof(unsigned char));
	unsigned char msg_tmp[m];
	memset(msg_tmp, 0, sizeof(msg_tmp));
	unsigned char tobyte_tmp[m];
	memset(tobyte_tmp, 0, m * sizeof(unsigned char));
	unsigned char input_tmp[2 * m];
	memset(input_tmp, 0, 2 * m * sizeof(unsigned char));
	toByte(tobyte_tmp, idx, m);
	concat_bytes(input_tmp, tobyte_tmp, m, r, m);
	Hm(msg_tmp, msg, msglen, input_tmp, m); /* byte[m] */

	/* byte[n] node = XMSS_rootFromSig(Sig', M, SEED, ADRS) */
	unsigned char node[n];
	memset(node, 0, sizeof(node));
	xmss_rootFromSig_wo_hash(&Sigp, msg_tmp, node, seed, m, h/d, adrs);
	for (unsigned int j = 1; j < d ; j++) {
		/* idx_leaf = (h / d) least significant bytes of idx_tree */
		getLSB(idx_tree, &idx_leaf, h / d, h);

		/*
		 * idx_tree = (h - j * h / d)
		 * most significant bytes of idx_tree
		 */
		getMSB(idx_tree, &idx_tree, h - j * (h / d), h);
		
		/* Sig' = idx_leaf || getXMSSSignature(Sig, j) */
		Sigp.idx_sig = idx_leaf;
		memcpy(Sigp.sig_ots, Sig_MT.reduced_sigs[j-1].sig_ots,
		    para->len * m * sizeof(*Sigp.sig_ots));
		memcpy(Sigp.auth, Sig_MT.reduced_sigs[j-1].auth,
		    (h / d) * m * sizeof(*Sigp.auth));

		setLayerAddress(adrs, j);	/* ADRS.setLayerAddress(j) */
		setTreeAddress(adrs, idx_tree);	/* ADRS.setTreeAddress(idx_tree) */

		xmss_rootFromSig_wo_hash(&Sigp, node, node, seed, m, h/d, adrs);
		/* node = XMSS_rootFromSig_wo_hash(Sig', node, SEED, ADRS); */
	}
	xmss_sig_free(&Sigp);
	xmss_mt_sig_free(&Sig_MT, d);

	int success = 1;
	if (memcmp(node, PK_MT.root, m) == 0)
		success = 0;
	xmss_pk_free(&PK_MT);
	return success; /* 0 if verification successful, else 1 */
}

void
xmss_mt_StringToSig(xmss_mt_sig *Sig, unsigned char *string, unsigned int m,
    unsigned int n, unsigned int h, unsigned int w, unsigned int d)
{
	unsigned int len1 = calcLen1(w, m);
	unsigned int len = calcLen(len1, calcLen2(w, len1));

	/* Extract idx_sig from string */
	unsigned int length_idx = ceil(h / 8.0), intermediate_length = 0;
	uint64_t idx = 0;

	for (unsigned int i = 0; i < length_idx; i++)
		idx |= (uint64_t)string[i] << 8 * (length_idx - 1 - i);

	Sig->idx_sig = idx;

	/* Extract randomness */
	memcpy(Sig->randomness, &string[length_idx],
	    m * sizeof(*Sig->randomness));
	intermediate_length = length_idx + m;

	/* Extract bottom_reduced_sig */
	memcpy(Sig->bottom_reduced_sig.sig_ots, &string[intermediate_length],
	    len * n * sizeof(*Sig->bottom_reduced_sig.sig_ots));
	intermediate_length += len * n;
	memcpy(Sig->bottom_reduced_sig.auth, &string[intermediate_length],
	    (h / d) * n * sizeof(*Sig->bottom_reduced_sig.auth));
	intermediate_length += (h / d) * n;

	/* Compute len_n */
	unsigned int len_n = calcLen(calcLen1(w, n),
			calcLen2(w, calcLen1(w, n)));

	/* Extract the array of d - 1 reduced signatures */
	for (unsigned int i = 0 ; i < d - 1; i++) {
		memcpy(Sig->reduced_sigs[i].sig_ots,
		    &string[intermediate_length],
		    len_n * n * sizeof(unsigned char));
		intermediate_length += len_n * n;
		memcpy(Sig->reduced_sigs[i].auth, &string[intermediate_length],
		    (h / d) * n * sizeof(unsigned char));
		intermediate_length += (h / d) * n;
	}
}

int
xmss_mt_SigToString(unsigned char *concat, xmss_mt_sig *Sig, unsigned int m,
    unsigned int n, unsigned int h, unsigned int w, unsigned int d)
{
	/* returns 1 if a problem with calloc occurred */
	unsigned int len1 = calcLen1(w, m);
	unsigned int len = calcLen(len1, calcLen2(w, len1));

	/* In our case len = len_n, otherwise use calc()-functions like above */
	unsigned int len_n = len;
	unsigned int length_idx = ceil(h / 8.0), intermediate_length = 0;
	uint64_t idx = Sig->idx_sig;
	unsigned char *idx_sig_char;
	if ((idx_sig_char = calloc(length_idx, sizeof(*idx_sig_char))) == NULL)
		return 1;
	for (unsigned int i = 0; i < length_idx; i++) {
		idx_sig_char[i] = (unsigned char)
		    (idx >> (8 * (length_idx - 1 - i))) & 0xFF;
	}
	concat_bytes(concat, idx_sig_char, length_idx, Sig->randomness, m);
	intermediate_length = m + length_idx;
	concat_bytes(concat, concat, intermediate_length,
			Sig->bottom_reduced_sig.sig_ots, len * n);
	intermediate_length += len * n;
	concat_bytes(concat, concat, intermediate_length,
			Sig->bottom_reduced_sig.auth, (h / d) * n);
	intermediate_length += (h / d) * n;
	for (unsigned int j = 0; j < d - 1; j++) {
		concat_bytes(concat, concat, intermediate_length,
		    Sig->reduced_sigs[j].sig_ots, len_n * n);
		intermediate_length += len_n * n;
		concat_bytes(concat, concat, intermediate_length,
		    Sig->reduced_sigs[j].auth, (h / d) * n);
		intermediate_length += (h / d) * n;
	}
	free(idx_sig_char);
	return 0;
}
